/*
 * VERSION 3.00
 *
 * REMARK December 10th 2021:
 * This version supports all features of the previous versions plus
 * an environment and benchmark for energy flux analysis with the VWPM (not for the scalar BPM or scalar WPM!)
 * Run `bash ./bin/consigure.sh <option>` where <option> is in {flux} to configure the environment
 *   max performance is achieved with "none"
 *   thread timing analysis is performed with "timer"
 *   most debug information is obtained with "debug"
 *   electromagnetic flux is calculated and analyzed with "flux"
 *
 * Please read the following webpage to understand the parameters supported by the 'flux' environment:
 *   http://www-home.htwg-konstanz.de/~MFERTIG/pages/proglab.html#energyflux
 *
 * Please review the following files to understand the energy flux benchmark
 *   ./bin/benchmark3D_flux/flux_benchmark_cpu.sh
 *   ./bin/benchmark3D_flux/vwpm/config*/flux.cfg
 *
 * Do the following to start the benchmark
 *   $ cd bin
 *   $ bash configure flux
 * (answer 'y' on the question)
 *   $ cd benchmark3D_flux
 *   $ bash flux_benchmark_cpu.sh
 *
 * UDPM - An electromagnetic Fourier simulator environment 
 *        for education and training purposes
 *        developed by Matthias Fertig since 2019
 *
 *        Please do NOT use, copy or distribute the environment in parts or as a whole for any commercial use!
 *        Please make a reference to the source and the author of this environment if used for non-commercial purpose.
 *        All rights reserverd.
 */

/*
 * VERSION 2.01
 *
 * REMARK July 29th 2021:
 * This version supports 2D and 3D simulations for scalar BPM and WPM
 *   2D and 3D simulations on a GPU are supported for the scalar BPM only
 *   2D and 3D simulations on a GPU for the scalar WPM are pending
 * Run `bash ./bin/consigure.sh <option>` where <option> is in {none, flux, debug, timer, all} to configure your runs
 *   max performance is achieved with "none"
 *   thread timing analysis is performed with "timer"
 *   most debug information is obtained with "debug"
 *   electromagnetic flux is calculated witth "flux"
 *
 * UDPM - An electromagnetic Fourier simulator environment 
 *        for education and training purposes
 *        developed by Matthias Fertig since 2019
 *
 *        Please do NOT use, copy or distribute the environment in parts or as a whole for any commercial use!
 *        All rights reserverd.
 */

/* ***********************
 *     General remarks
 * ***********************/

This environment supports Beam and Wave Propagation Methods for electromagnetic scalar and vector waves. It is part of a university course where students are trained in programming of electromagnetic Fourier simulators and massively parallel systems, i.e. Graphics Processing Units. The environment contains reference code for each of the supported methods in a static library. Students provide their code in the your_pm.cc file and then compile an executable against the static library. The executable reads a configuration file (e.g. wguide_bpm_2D.cfg), performs a simulation of the student code and then reports on simulation error and runtime. Students get an immediate feedback about the correctness and performance of their codes, compared to the reference code in the static library.



/* ***********************
 *     FFT library
 * ***********************/

The reference code utilizes the FFTW library to perform Fast Fourier Transformations



/* ******************
 *     Contents
 * ******************/

/* CPU applications */

./src/your_pm.cc	- student reference code fpr CPU applications
./inc/your_pm.hh  	- header file of studenc reference code for CPU applications
		   	(ATTENTION! Do not modify the interfaces of pre-defined functions as they are called by the code in the static library provided with the environment

/* GPU applications */
./src/your_kernel.cu	- student reference code fpr GPU applications
./inc/your_kernel.hh  	- header file of studenc reference code for CPU applications
		   	(ATTENTION! Do not modify the interfaces of pre-defined functions as they are called by the code in the static library provided with the environment

./lib_udpmV2.01a 	- static library with reference code
./Makefile		- 'make' compiles the new executable against the static library
./bin/pm		- executable

			Example:
			$cd bin
			$./pm -i wguide_bpm_2D.txt -c -p -r qor.csv

			Open PROF_E.csv and color-code region to inspect electric field distribution
			Open PROF_F.csv and color-code region to inspect energy flux



/* ***********************
 *     Programming Guide
 * ***********************/

			/* **************************
			 * Student CPU applications 
			 * **************************/
./src/your_pm.cc

/* 3D vectorial methods */
void YOUR_vectorial_3D(cpx* EX, cpx* EY, cpx* EZ, double* F, 
		       const cpx* N, const cpx* KX, const double* X, const cpx* KY, const double* Y, 
		       const double k0, const double dz, const double dy, const double dx, 
		       const int nz, const int ny, const int nx, 
		       const bool do_evanescent_modes = 1);
/* 2D vectorial methods */
void YOUR_vectorial_2D(cpx* EX, cpx* EY, cpx* EZ, double* F, 
		       const cpx* N, const cpx* KX, const double* X,
		       const double k0, const double dz, const double dx, 
		       const int nz, const int nx, 
		       const bool do_evanescent_modes = 1);
/* 3D scalar methods */
void YOUR_scalar_3D(cpx* E, double* F, 
		    const cpx* N, const cpx* KX, const double* X, const cpx* KY, const double* Y, 
		    const double k0, const double dz, const double dy, const double dx, 
		    const int nz, const int ny, const int nx, 
		    const bool do_evanescent_modes = 1, const bool is_TE = true);
/* 2D scalar methods */
void YOUR_scalar_2D(cpx* E, double* F,
		    const cpx* N, const cpx* KX, const double* X,
		    const double k0, const double dz, const double dx,
		    const int nz, const int nx,
		    const bool do_evanescent_modes = 1, const bool is_TE = true);


		    /* *************************
 		     * Student GPU applications
 		     * *************************/

./src/your_kernel.cu

/* 3D vectorial methods */
void YOUR_kernelWrapper_3D( int blockSize,
                            cufftDoubleComplex* EX, cufftDoubleComplex* EY, cufftDoubleComplex* EZ, double* F, 
                            cufftDoubleComplex* N, 
			    cufftDoubleComplex* KX, double* X, 
			    cufftDoubleComplex* KY, double* Y, 
                            double k0, double dz, double dy, double dx, int nz, int ny, int nx, 
                            bool do_evanescent_modes = false);
/* 3D scalar methods */
void YOUR_kernelWrapper_3D( int blockSize,
                            cufftDoubleComplex* E, double* F, 
                            cufftDoubleComplex* N, 
			    cufftDoubleComplex* KX, double* X, 
			    cufftDoubleComplex* KY, double* Y, 
                            double k0, double dz, double dy, double dx, int nz, int ny, int nx, 
                            bool do_evanescent_modes = false, bool is_TE = true );
/* 2D vectorial */
void YOUR_kernelWrapper_2D( int blockSize,
                            cufftDoubleComplex* EX, cufftDoubleComplex* EY, cufftDoubleComplex* EZ, double* F, 
			    cufftDoubleComplex* N, 
			    cufftDoubleComplex* KX, double* X, 
			    double k0, double dz, double dx, int nz, int nx, 
			    bool do_evanescent_modes = false );
/* 2D scalar */
void YOUR_kernelWrapper_2D( int blockSize,
			    cufftDoubleComplex* E, double* F, 
     			    cufftDoubleComplex* N, 
			    cufftDoubleComplex* KX, double* X, 
			    double k0, double dz, double dx, int nz, int nx, 
			    bool do_evanescent_modes = false, bool is_TE = true );



			/* **************************
			 * Compile executable
			 * **************************/

$ pwd
$ ./bin
$ cd ..
$ make
$ cd bin
$ ls -l pm
$ -rwxrwxr-x 1 student student 3457379 <actual date and time>



			/* **************************
			 * Run examle wguide_bpm_2D.txt
			 * **************************/

$ pwd
$ ./bin
$ ./pm -i wguide_bpm_2D.cfg -c -p -r qor.csv



/* ****************
 *     Output:
 * ***************/
$ ./pm -i wguide_bpm_2D.txt -c -p -r qor.csv 


	Options ... 
		option -r : --store-results qor.csv
		option -p : --run-cpu
		option -g : --run-gpu
		option -c : --compare
		option -i : --config-file wguide_bpm_2D.cfg
		option -k : --kernel 0
		option -b : --block-size 16
	Options ... 		done.

	Configuration file ...		wguide_bpm_2D.cfg
	Parsing ... 		wguide_bpm_2D.cfg
				[global]
					nx =	48
					ny =	0
					nz =	480
					x =	2000
					y =	0
					z =	20000
					evanescent =	0
					scale =	nm
					method =	bpm
					pol =	TE
				[/global]
				[layer]
					type =	waveguide
					x =	0
					y =	0
					z =	20000
					r =	250
					n =	3.4
					k =	0
					ne =	1.0
					ke =	0.02
					dx =	0
					dy =	0
					dz =	0
				[/layer]
				[wave]
					shape =	gaussian
					sigmax =	150
					sigmay =	150
					x =	0
					y =	0
					z =	0
					theta =	0
					phi =	0
					beta =	0
					amp =	1
					lambda =	850
				[/wave]
	Parsing ... 		done.

	Initialization ...
		EvModes ...	0
		Method ...	bpm
	Initialization ... 	 done.

	Sampling ...
		Array size ...	nx*ny*nz = 23040 entries.
		Space X ...	done.	nx=48, dx=4.16667e-08m, ax=2e-06m
		Space Z ...	done.	nz=480, dz=4.16667e-08m, az=2e-05m
		Freq X ...	done.	dfx=0.5MHz, fmax_x=24MHz, flim_x=12MHz, dtheta_x=1.19375°
	Sampling ... 		done.

	System ...
				create WGUIDE ... done.
	System ... 		done.

	Energy flux ... 	
	                	(STUD_) memory allocated.
	                	(PROF_) memory allocated.
	Energy flux ... 	done.


	Wave ...
				create GAUSSIAN ...

				scale = 1e-09
				amplitude = 1
				lambda = 850
				theta = 0
				x = 0
				z = 0
				sx = 0
				sz = 1

	         		lambda = 8.5e-07

	         		option -c (--compare) causes duplication of memory for comparison
	Wave ... 		done.

Initialization ... 		done. 2.13215ms

	Output ... 		done.
	           		option -x (--cross-section) is -1
	           		option -l (--layer) is 0

	System ... 		export to N.csv done.


Applying simulation on GPU ...
	Memory allocation ...	done. 0.391021ms
	Kernel call ...		done. 367.862ms
	Copy data D2H ...	done. 0.247402ms
	GPU flux ... 		@z=0 is 8.84119e+08
	GPU run ... 		completed (PROF_)
	            		gpu runtime 367.862 ms

Applying simulation on GPU ...
	Memory allocation ...	done. 0.35257ms
	Kernel call ...		done. 38.1157ms
	Copy data D2H ...	done. 0.218001ms
	GPU flux ... 		@z=0 is 8.84119e+08
	GPU run ... 		completed (STUD_)
	            		gpu runtime 38.1157 ms
Error analysis ... 
	max error ... 		E [ max( (PROF__gpu(x,y,z)-STUD__gpu(x,y,z))/PROF__gpu(x,y,z) ) ]: 0

	[32mCONGRATULATIONS, maximum gpu-error is within tolerance of 0.1 promille![0m

GPU runtime analysis ... 
	            		(PROF_) gpu runtime 367.862 ms
	            		(STUD_) gpu runtime 38.1157 ms
	            		delta (absolute) = 	1e3*(PROF_time-STUD_time) = 		329.746 ms
	            		delta (relative) = 	1e2*(PROF_time-STUD_time)/PROF_time = 	89.6386 %
	            		delta (factor)   = 	PROF_time/STUD_time = 			9.6512


Applying simulation on CPU ... 
	CPU flux ... 		@z=0 is 5.91151e+07

	CPU run ... 		completed (PROF_)
	            		cpu runtime 7.89617 ms

Applying simulation on CPU ... your_pm.cc: function YOUR_scalar_2D()
	CPU flux ... 		@z=0 is 5.91151e+07

	CPU run ... 		completed (STUD_)
	            		cpu runtime 5.23856 ms


CPU correctness analysis ... 
	max Error E [ max( (PROF_cpu(x,y,z)-STUD_cpu(x,y,z))/PROF_cpu(x,y,z) ) ]: 0

	[32mCONGRATULATIONS, maximum cpu-error is within tolerance of 0.1 promille![0m

CPU runtime analysis ... 
	            		(PROF_) cpu runtime 7.89617 ms
	            		(STUD_) cpu runtime 5.23856 ms
	            		delta (absolute) = 	1e3*(PROF_time-STUD_time) = 	2.65761 ms
	            		delta (relative) = 	(PROF_time-STUD_time)/PROF_time*100 = 	33.657 %
	            		delta (factor) = 	PROF_time/STUD_time = 	1.50732



			/* **************************
			 * Configuration file
			 * **************************/

[global]
nx = 48
ny = 0
nz = 480
x = 2000
y = 0
z = 20000
evanescent = 0
scale = nm
method = bpm
polarization = TE
[/global]

[layer]
type = waveguide
x = 0
y = 0
z = 20000
r = 250
n = 3.4
k = 0
ne = 1.0
ke = 0.02
dx = 0
dy = 0
dz = 0
[/layer]

[wave]
shape = gaussian
sigmax = 150
sigmay = 150
x = 0
y = 0
z = 0
theta = 0
phi = 0
beta = 0
amp = 1
lambda = 850
[/wave]


